/* pong-i18n-tool: pong i18n tool
 * Author: George Lebl
 * (c) 1999 the Free Software Foundation
 * (c) 2000 Eazel, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "config.h"
#include <string.h>
#include <gnome.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>
#include <gnome-xml/xmlmemory.h>

#include "../pong/pong.h"
#include "../pong/pongpane.h"
#include "../pong/pongparser.h"
#include "../pong/pongutil.h"
#include "../pong/pongtrans.h"
#include "../pong/pong-strings.h"

static PongFile *pong_file = NULL;

static gboolean noerror = FALSE;

static const struct poptOption options[] = {
	{ "noerror", 0, POPT_ARG_NONE, &noerror, 0, N_("Don't die on errors, if you can help it"), N_("NOERROR") },
	{ NULL } 
};

typedef struct {
	char *string;
	char *translation;
	gboolean used;
} Translation;

static GHashTable *translations = NULL;

static Translation *
add_translation (const char *string,
		 const char *translation,
		 gboolean used)
{
	Translation *t = g_new0 (Translation, 1);

	t->string = g_strdup (string);
	t->translation = g_strdup (translation);
	t->used = used;

	/* A leak? (on duplicates), yes but who cares */
	g_hash_table_insert (translations, t->string, t);

	return t;
}

static Translation *
get_translation (const char *string)
{
	Translation *t;

        t = g_hash_table_lookup (translations, string);

	if (t != NULL)
		return t;

	t = add_translation (string, NULL, FALSE);

	return t;
}

static void
touch_string (const char *string)
{
	Translation *t;

	if (string == NULL)
		return;

	t = get_translation (string);
	t->used = TRUE;
}

static void
touch_plug (PongPlugWidget *widget)
{
	touch_string (widget->base.label);
}

static void
touch_widget (PongWidget *widget)
{
	GList *li;

	touch_string (widget->base.label);
	touch_string (widget->tooltip);

	for (li = widget->arguments; li != NULL; li = li->next) {
		PongArgument *arg = li->data;

		if (arg->value_translatable)
			touch_string (arg->value);
	}

	for (li = widget->options; li != NULL; li = li->next) {
		PongOption *opt = li->data;

		touch_string (opt->label);
	}
}

static void
touch_group (PongGroup *group)
{
	GList *li;

	touch_string (group->base.label);

	for (li = group->widgets; li != NULL; li = li->next) {
		PongWidget *widget = li->data;

		if (widget->base.type == PONG_PLUG_WIDGET)
			touch_plug ((PongPlugWidget *)widget);
		else
			touch_widget (widget);
	}
}

static void
touch_pane (PongPane *pane)
{
	GList *li;

	touch_string (pane->base.label);

	for (li = pane->groups; li != NULL; li = li->next) {
		PongGroup *group = li->data;

		if (group->base.type == PONG_PLUG_WIDGET)
			touch_plug ((PongPlugWidget *)group);
		else
			touch_group (group);
	}
}

static void
touch_file (PongFile *file)
{
	GList *li;

	touch_string (pong_file->dialog_title);
	touch_string (pong_file->help_path);

	for (li = file->panes; li != NULL; li = li->next) {
		PongPane *pane = li->data;
		touch_pane (pane);
	}
}

static void
read_translation (xmlNodePtr parent, gboolean used)
{
	xmlNodePtr node;

	char *string = NULL;
	char *translation = NULL;

	for (node = parent->childs; node != NULL; node = node->next) {
		/* possible leak on bad (with dupes) files, boohoo */
		if (pong_strcasecmp_no_locale (node->name, PONG_S_String) == 0) {
			string = xmlNodeGetContent (node);
		} else if (pong_strcasecmp_no_locale (node->name, PONG_S_Translated) == 0) {
			translation = xmlNodeGetContent (node);
		}
	}

	if (string != NULL)
		add_translation (string, translation, used);

	if (string != NULL)
		xmlFree (string);

	if (translation != NULL)
		xmlFree (translation);
}

static void
read_translations (xmlDocPtr doc, gboolean mark_unused)
{
	xmlNodePtr node;

	for (node = doc->root->childs; node != NULL; node = node->next) {
		if (pong_strcasecmp_no_locale (node->name, PONG_S_Translation) == 0) {
			if (mark_unused)
				read_translation (node, FALSE /*used*/);
			else
				read_translation (node, TRUE /*used*/);
		} else if (pong_strcasecmp_no_locale (node->name, PONG_S_UnusedTranslation) == 0) {
			read_translation (node, FALSE /*used*/);
		}
	}
}

static void
add_translation_to_doc (xmlDocPtr doc, Translation *t)
{
	xmlNodePtr node;

	if (t->used)
		node = xmlNewChild (doc->root, NULL, PONG_S_Translation, NULL);
	else
		node = xmlNewChild (doc->root, NULL, PONG_S_UnusedTranslation, NULL);
	xmlNewChild (node, NULL, PONG_S_String, t->string);
	if (t->translation != NULL)
		xmlNewChild (node, NULL, PONG_S_Translated, t->translation);
	else
		xmlNewChild (node, NULL, PONG_S_Translated, "");
}


static void
add_used_translations (gpointer key, gpointer value, gpointer user_data)
{
	xmlDocPtr doc = user_data;
	Translation *t = value;

	if (t->used)
		add_translation_to_doc (doc, t);
}

static void
add_unused_translations (gpointer key, gpointer value, gpointer user_data)
{
	xmlDocPtr doc = user_data;
	Translation *t = value;

	if ( ! t->used)
		add_translation_to_doc (doc, t);
}

static xmlDocPtr
write_translations (const char *locale)
{
	xmlDocPtr doc;
	xmlNodePtr newxml;

	doc = xmlNewDoc ("1.0");
	doc->root = xmlNewDocNode (doc, NULL, PONG_S_PongTranslations, NULL);

	newxml = xmlNewChild (doc->root, NULL, PONG_S_Locale, locale);

	g_hash_table_foreach (translations, add_used_translations, doc);
	g_hash_table_foreach (translations, add_unused_translations, doc);

	return doc;
}

int
main (int argc, char *argv[])
{
	xmlDocPtr doc;
	poptContext ctx;
	const char **args;
	const char *file, *locale;
	char *found_file, *transl_file;

	gnomelib_init ("pong-i18n-tool", VERSION);

	ctx = gnomelib_parse_args (argc, argv, 0);

	if ( ! pong_init ())
		g_error (_("PonG init failed!"));

	args = poptGetArgs (ctx);

	if (args == NULL ||
	    args[0] == NULL ||
	    args[1] == NULL ||
	    args[2] != NULL) {
		g_warning (_("You must supply two arguments, "
			     "the .pong file and the locale."));
		exit (1);
	}

	file = args[0];
	locale = args[1];

	found_file = pong_find_file (file, NULL);
	if (found_file == NULL) {
		g_warning (_("Cannot find %s"), file);
		exit (1);
	}

	/* Find and check for the translation file */
	transl_file = pong_make_translation_file (found_file, locale);

	g_print (_("Locale: %s\n"
		   "Pong file: %s\n"
		   "Translation file: %s\n"),
		 locale, found_file, transl_file);

	translations = g_hash_table_new (g_str_hash, g_str_equal);

	if (g_file_exists (transl_file)) {
		g_print (_("Locale file exists, will merge.\n"));

		doc = xmlParseFile (transl_file);

		if (doc != NULL &&
		    doc->root != NULL &&
		    doc->root->name != NULL &&
		    pong_strcasecmp_no_locale(doc->root->name, PONG_S_PongTranslations) == 0) {
			read_translations (doc, TRUE);
		} else {
			g_warning (_("File %s exists, but is not a PonG Translation file, aborting"));
			if ( ! noerror)
				exit (1);
		}
		xmlFreeDoc(doc);

		/* FIXME: This is an utter hack for backups */
		{
			char *foocmd = g_strdup_printf ("cp -f '%s' '%s.bak'",
							transl_file, transl_file);
			/* sanity check to make sure the above doesn't do something bad */
			if (strchr (transl_file, '\'') != NULL) {
				/* yaikes, but "oh well" */
				g_warning (_("No backup made, filename %s contains a signle quote"));
				if ( ! noerror)
					exit (1);
			} else {
				/* eeek! */
				g_print (_("Running: %s\n"), foocmd);
				if (system (foocmd) != 0) {
					g_warning (_("Failed to make backup!"));
					if ( ! noerror)
						exit (1);
				}
			}

			g_free (foocmd);
		}
	}
				
	pong_file = pong_file_load_xml (found_file,
					TRUE /*no_i18n*/,
					FALSE /*use_gettext*/,
					NULL /*gettext_domain*/,
					TRUE /*bogus_types*/);

	if (pong_file == NULL) {
		g_warning (_("Can't load pong file: %s"), found_file);
		exit (1);
	}

	g_print (_("Getting translations..."));

	touch_file (pong_file);

	g_print (_("\nWriting translations to %s..."), transl_file);

	doc = write_translations (locale);

	/* try to save the file */
	if (xmlSaveFile (transl_file, doc) == -1) {
		g_print ("\n");
		g_warning ("Can't save file %s", transl_file);
		exit (1);
	}

	xmlFreeDoc (doc);

	g_print (_("\nDone!\n"));

	return 0;
}
