/* PonG: the parser using SAX interface
 * Author: George Lebl
 * (c) 1999 Free Software Foundation
 * (c) 2000 Eazel, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "config.h"
#include <gnome.h>
#include <gnome-xml/parser.h>
#include <gnome-xml/parserInternals.h>

#include "pongpane.h"
#include "pongelement.h"
#include "pongutil.h"
#include "pongtrans.h"
#include "pong-strings.h"

#include "pongparser.h"

extern GList *pong_directories;

enum {
	IGNORE,
	START,
	FINISH,
	LOOKING_FOR_ELEMENT,

	INSIDE_ELEMENT,

	INSIDE_BASECONFPATH,
	INSIDE_DIALOG_TITLE,
	INSIDE_AUTO_APPLY,
	INSIDE_REVERT,
	INSIDE_HELP,
	INSIDE_HELP_NAME,
	INSIDE_HELP_PATH,
	INSIDE_LEVEL,
	INSIDE_LEVEL_NAME,
	INSIDE_LEVEL_BASECONFPATH,

	INSIDE_ELEMENT_TYPE,
	INSIDE_ELEMENT_CONF_PATH,
	INSIDE_ELEMENT_WIDGET,
	INSIDE_ELEMENT_DEFAULT,
	INSIDE_ELEMENT_SPECIFIER,

	INSIDE_SENSITIVITY,
	INSIDE_SENSITIVITY_CONNECTION,
	INSIDE_SENSITIVITY_VALUE,
	INSIDE_SENSITIVITY_COMPARISON,
	INSIDE_SENSITIVITY_SENSITIVE,
	INSIDE_SENSITIVITY_INSENSITIVE,

	INSIDE_PANE,
	INSIDE_PANE_NAME,
	INSIDE_PANE_LABEL,
	INSIDE_PANE_LEVEL,

	INSIDE_GLADE,
	INSIDE_GLADE_NAME,
	INSIDE_GLADE_LEVEL,
	INSIDE_GLADE_EXPAND,
	INSIDE_GLADE_FILE,
	INSIDE_GLADE_WIDGET,

	INSIDE_BONOBO,
	INSIDE_BONOBO_NAME,
	INSIDE_BONOBO_LEVEL,
	INSIDE_BONOBO_EXPAND,
	INSIDE_BONOBO_QUERY,

	INSIDE_PLUG,
	INSIDE_PLUG_NAME,
	INSIDE_PLUG_LEVEL,
	INSIDE_PLUG_EXPAND,
	INSIDE_PLUG_TYPE,
	INSIDE_PLUG_PATH,
	INSIDE_PLUG_SPECIFIER,

	INSIDE_GROUP,
	INSIDE_GROUP_NAME,
	INSIDE_GROUP_LABEL,
	INSIDE_GROUP_LEVEL,
	INSIDE_GROUP_EXPAND,

	INSIDE_WIDGET,
	INSIDE_WIDGET_NAME,
	INSIDE_WIDGET_LABEL,
	INSIDE_WIDGET_LEVEL,
	INSIDE_WIDGET_EXPAND,
	INSIDE_WIDGET_TYPE,
	INSIDE_WIDGET_TOOLTIP,

	INSIDE_ARGUMENT,
	INSIDE_ARGUMENT_NAME,
	INSIDE_ARGUMENT_VALUE,
	INSIDE_ARGUMENT_TRANSLATABLE,

	INSIDE_OPTION,
	INSIDE_OPTION_LABEL,
	INSIDE_OPTION_VALUE
};


typedef struct _State State;
struct _State {
	/* stuff that is read */
	PongFile *gf;
	PongSensitivity *sens;

	/* how the information is parsed */
	PongTranslation *trans;
	gboolean use_gettext;
	char *gettext_domain;
	gboolean bogus_types;

	int state;
	int ignore_last_state; /* state to go into after ignore */
	int ignore_depth; /* how deep are we into ingore */
	GString *buf;
	gboolean reading_characters;
	PongElement *element;

	PongPane *pane;
	PongGroup *group;
	PongWidget *widget;
	PongPlugWidget *plug_widget;

	int state_before_plug; /* state to go into after a plug */

	char *arg_name;
	char *arg_value;
	gboolean arg_translatable;

	char *opt_label;
	char *opt_value;

	PongLevel *level;
};

static void startDocument		(gpointer user_data);
static void endDocument			(gpointer user_data);
static void characters			(gpointer user_data,
					 const CHAR *ch,
					 int len);
static void startElement		(gpointer user_data,
					 const CHAR *name,
					 const CHAR **attrs);
static void endElement			(gpointer user_data,
					 const CHAR *name);

static int myXmlSAXParseFile		(xmlSAXHandlerPtr sax,
					 gpointer user_data,
					 const char *filename);
static int myXmlSAXParseMemory		(xmlSAXHandlerPtr sax,
					 gpointer user_data,
					 const char *buffer,
					 int size);

static void pong_file_add_element	(PongFile *gf,
					 PongElement *ge);
static void pong_file_destroy		(PongFile *gf);

static void unref_elements		(gpointer key,
					 gpointer data,
					 gpointer user_data);
static gboolean kill_unused_pongfiles	(gpointer key,
					 gpointer data,
					 gpointer user_data);
static void add_elements_to_list	(gpointer key,
					 gpointer data,
					 gpointer user_data);
static void foreach_elements		(gpointer key,
					 gpointer data,
					 gpointer user_data);

static PongComparison get_comparison	(const char *name);

static xmlSAXHandler handler = {0};
static GHashTable *file_hash = NULL;
static gboolean parser_inited = FALSE;

static GHashTable *comparison_hash = NULL;

#include "pong-comparisons.cP"

static void
init_comparison_hash (void)
{
	int i;
	if (comparison_hash != NULL)
		return;

	comparison_hash = g_hash_table_new (g_str_hash, g_str_equal);

	for (i = 0; comparisons[i].name != NULL; i++) {
		g_hash_table_insert (comparison_hash, comparisons[i].name,
				     GINT_TO_POINTER (comparisons[i].value));
	}
}

static PongComparison
get_comparison (const char *name)
{
	g_return_val_if_fail (name != NULL, PONG_COMP_INVALID);

	if (comparison_hash == NULL)
		init_comparison_hash ();

	/* INVALID == 0, so if it can't find it it will
	 * return INVALID */
	return GPOINTER_TO_INT (g_hash_table_lookup (comparison_hash, name));
}

static PongConnection
get_connection (const char *name)
{
	g_return_val_if_fail (name != NULL, PONG_CONNECTION_INVALID);

	if (pong_strcasecmp_no_locale (name, PONG_S_OR) == 0)
		return PONG_CONNECTION_OR;
	else if (pong_strcasecmp_no_locale (name, PONG_S_AND) == 0)
		return PONG_CONNECTION_AND;
	else
		return PONG_CONNECTION_INVALID;
}

static const char *
translate_string (State *state, const char *string)
{
	if (state->use_gettext) {
#ifdef ENABLE_NLS
		if (state->gettext_domain != NULL)
			return dgettext (state->gettext_domain, string);
		else
			return gettext (string);
#else
		return string;
#endif
	} else {
		return pong_translation_translate (state->trans, string);
	}
}


static void
parser_init (void)
{
	if (parser_inited)
		return;

	handler.startDocument = startDocument;
	handler.endDocument = endDocument;
	handler.characters = characters;
	handler.startElement = startElement;
	handler.endElement = endElement;

	file_hash = g_hash_table_new (g_str_hash, g_str_equal);

	parser_inited = TRUE;
}

static void
startDocument (gpointer user_data)
{
	State *state = user_data;

	state->state = START;

	state->buf = g_string_new ("");
	state->reading_characters = FALSE;
}

static void
endDocument (gpointer user_data)
{
	State *state = user_data;

	if (state->sens != NULL) {
		pong_sensitivity_destroy (state->sens);
		state->sens = NULL;
	}

	if (state->buf != NULL) {
		g_string_free (state->buf, TRUE);
		state->buf = NULL;
	}
	state->reading_characters = FALSE;
}

static void
characters (gpointer user_data, const CHAR *ch, int len)
{
	State *state = user_data;
	char *s;

	if ( ! state->reading_characters)
		return;

	/* there must be a better way of doing this */
	s = g_strndup (ch, len);
	g_string_append (state->buf, s);
	g_free (s);
}

#define START_READING_CHARS(element) 			\
	state->state = element;				\
							\
	/* clear buffer for attribute */		\
	g_string_assign (state->buf, "");		\
							\
	/* read characters now */			\
	state->reading_characters = TRUE;

#define START_IGNORE 					\
	state->ignore_last_state = state->state;	\
	state->state = IGNORE;				\
	state->ignore_depth = 1;

static void
startElement (gpointer user_data, const CHAR *name, const CHAR **attrs)
{
	State *state = user_data;

	switch (state->state) {
	case IGNORE:
		state->ignore_depth++;
		break;

	case START:
		if (pong_strcasecmp_no_locale (name, PONG_S_PongElements) == 0) {
			state->state = LOOKING_FOR_ELEMENT;
		} else {
			START_IGNORE;
		}
		break;

	case FINISH: /* just ingore everything now */
		break;

	case LOOKING_FOR_ELEMENT:
		if (pong_strcasecmp_no_locale (name, PONG_S_Element) == 0) {
			state->state = INSIDE_ELEMENT;
			state->element = pong_element_new ();
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Pane) == 0) {
			state->state = INSIDE_PANE;
			state->pane = pong_pane_new ();
		} else if (pong_strcasecmp_no_locale (name, PONG_S_BaseConfPath) == 0) {
			START_READING_CHARS (INSIDE_BASECONFPATH)
		} else if (pong_strcasecmp_no_locale (name, PONG_S_DialogTitle) == 0) {
			START_READING_CHARS (INSIDE_DIALOG_TITLE)
		} else if (pong_strcasecmp_no_locale (name, PONG_S_AutoApply) == 0) {
			START_READING_CHARS (INSIDE_AUTO_APPLY)
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Revert) == 0) {
			START_READING_CHARS (INSIDE_REVERT)
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Help) == 0) {
			state->state = INSIDE_HELP;
			g_free (state->gf->help_name);
			state->gf->help_name = g_strdup ("pong");
			g_free (state->gf->help_path);
			state->gf->help_path = g_strdup ("generic.html");
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			state->level = g_new0 (PongLevel, 1);
			state->state = INSIDE_LEVEL;
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_HELP:
		if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_HELP_NAME)
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Path) == 0) {
			START_READING_CHARS (INSIDE_HELP_PATH)
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_LEVEL:
		if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_LEVEL_NAME)
		} else if (pong_strcasecmp_no_locale (name, PONG_S_BaseConfPath) == 0) {
			START_READING_CHARS (INSIDE_LEVEL_BASECONFPATH)
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_ELEMENT:
		if (pong_strcasecmp_no_locale (name, PONG_S_Type) == 0) {
			START_READING_CHARS (INSIDE_ELEMENT_TYPE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_ConfPath) == 0) {
			START_READING_CHARS (INSIDE_ELEMENT_CONF_PATH);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Widget) == 0) {
			START_READING_CHARS (INSIDE_ELEMENT_WIDGET);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Default) == 0) {
			START_READING_CHARS (INSIDE_ELEMENT_DEFAULT);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Specifier) == 0) {
			START_READING_CHARS (INSIDE_ELEMENT_SPECIFIER);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Sensitivity) == 0) {
			if (state->element->type == PONG_TYPE_INVALID) {
				g_warning (_("Type not set before Sensitivity"));
				state->ignore_last_state = state->state;
				state->state = IGNORE;
				state->ignore_depth = 1;
			} else  {
				state->state = INSIDE_SENSITIVITY;
				state->sens =
					pong_sensitivity_new (state->element);
			}
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_SENSITIVITY:
		if (pong_strcasecmp_no_locale (name, PONG_S_Value) == 0) {
			START_READING_CHARS (INSIDE_SENSITIVITY_VALUE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Connection) == 0) {
			START_READING_CHARS (INSIDE_SENSITIVITY_CONNECTION);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Comparison) == 0) {
			START_READING_CHARS (INSIDE_SENSITIVITY_COMPARISON);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Sensitive) == 0) {
			START_READING_CHARS (INSIDE_SENSITIVITY_SENSITIVE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Insensitive) == 0) {
			START_READING_CHARS (INSIDE_SENSITIVITY_INSENSITIVE);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_PANE:
		if (pong_strcasecmp_no_locale (name, PONG_S_Group) == 0) {
			state->state = INSIDE_GROUP;
			state->group = pong_group_new ();
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Plug) == 0) {
			state->state = INSIDE_PLUG;
			state->state_before_plug = INSIDE_PANE;
			state->plug_widget = pong_plug_widget_new ();
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Glade) == 0) {
			state->state = INSIDE_GLADE;
			state->state_before_plug = INSIDE_PANE;
			state->plug_widget = pong_plug_widget_new ();
			pong_plug_widget_set_plug_type (state->plug_widget, PONG_S_Glade);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Bonobo) == 0) {
			state->state = INSIDE_BONOBO;
			state->state_before_plug = INSIDE_PANE;
			state->plug_widget = pong_plug_widget_new ();
			pong_plug_widget_set_plug_type (state->plug_widget, PONG_S_Bonobo);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_PANE_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Label) == 0) {
			START_READING_CHARS (INSIDE_PANE_LABEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			START_READING_CHARS (INSIDE_PANE_LEVEL);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_PLUG:
		if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_PLUG_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			START_READING_CHARS (INSIDE_PLUG_LEVEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Expand) == 0) {
			START_READING_CHARS (INSIDE_PLUG_EXPAND);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Type) == 0) {
			START_READING_CHARS (INSIDE_PLUG_TYPE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Path) == 0) {
			START_READING_CHARS (INSIDE_PLUG_PATH);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Specifier) == 0) {
			START_READING_CHARS (INSIDE_PLUG_SPECIFIER);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_GLADE:
		if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_GLADE_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			START_READING_CHARS (INSIDE_GLADE_LEVEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Expand) == 0) {
			START_READING_CHARS (INSIDE_GLADE_EXPAND);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_File) == 0) {
			START_READING_CHARS (INSIDE_GLADE_FILE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Widget) == 0) {
			START_READING_CHARS (INSIDE_GLADE_WIDGET);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_BONOBO:
		if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_BONOBO_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			START_READING_CHARS (INSIDE_BONOBO_LEVEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Expand) == 0) {
			START_READING_CHARS (INSIDE_BONOBO_EXPAND);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Query) == 0) {
			START_READING_CHARS (INSIDE_BONOBO_QUERY);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_GROUP:
		if (pong_strcasecmp_no_locale (name, PONG_S_Widget) == 0) {
			state->state = INSIDE_WIDGET;
			state->widget = pong_widget_new ();
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Glade) == 0) {
			state->state = INSIDE_GLADE;
			state->state_before_plug = INSIDE_GROUP;
			state->plug_widget = pong_plug_widget_new ();
			pong_plug_widget_set_plug_type (state->plug_widget, PONG_S_Glade);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Bonobo) == 0) {
			state->state = INSIDE_BONOBO;
			state->state_before_plug = INSIDE_GROUP;
			state->plug_widget = pong_plug_widget_new ();
			pong_plug_widget_set_plug_type (state->plug_widget, PONG_S_Bonobo);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_GROUP_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Label) == 0) {
			START_READING_CHARS (INSIDE_GROUP_LABEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			START_READING_CHARS (INSIDE_GROUP_LEVEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Expand) == 0) {
			START_READING_CHARS (INSIDE_GROUP_EXPAND);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_WIDGET:
		if (pong_strcasecmp_no_locale (name, PONG_S_Argument) == 0) {
			state->state = INSIDE_ARGUMENT;
			state->arg_name = NULL;
			state->arg_value = NULL;
			state->arg_translatable = FALSE;
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Option) == 0) {
			state->state = INSIDE_OPTION;
			state->opt_label = NULL;
			state->opt_value = NULL;
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_WIDGET_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Label) == 0) {
			START_READING_CHARS (INSIDE_WIDGET_LABEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Level) == 0) {
			START_READING_CHARS (INSIDE_WIDGET_LEVEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Expand) == 0) {
			START_READING_CHARS (INSIDE_WIDGET_EXPAND);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Type) == 0) {
			START_READING_CHARS (INSIDE_WIDGET_TYPE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Tooltip) == 0) {
			START_READING_CHARS (INSIDE_WIDGET_TOOLTIP);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_ARGUMENT:
		if (pong_strcasecmp_no_locale (name, PONG_S_Name) == 0) {
			START_READING_CHARS (INSIDE_ARGUMENT_NAME);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Value) == 0) {
			START_READING_CHARS (INSIDE_ARGUMENT_VALUE);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Translatable) == 0) {
			START_READING_CHARS (INSIDE_ARGUMENT_TRANSLATABLE);
		} else {
			START_IGNORE;
		}
		break;

	case INSIDE_OPTION:
		if (pong_strcasecmp_no_locale (name, PONG_S_Label) == 0) {
			START_READING_CHARS (INSIDE_OPTION_LABEL);
		} else if (pong_strcasecmp_no_locale (name, PONG_S_Value) == 0) {
			START_READING_CHARS (INSIDE_OPTION_VALUE);
		} else {
			START_IGNORE;
		}
		break;

	default:
		START_IGNORE;
		break;
	}
}

static void
endElement (gpointer user_data, const CHAR *name)
{
	State *state = user_data;
	int i, len;

	switch (state->state) {
	case IGNORE:
		state->ignore_depth--;
		if (state->ignore_depth == 0)
			state->state = state->ignore_last_state;
		break;

	case FINISH: /* just ingore everything now */
		break;

	case LOOKING_FOR_ELEMENT:
		state->state = FINISH;
		break;

	case INSIDE_ELEMENT:
		state->state = LOOKING_FOR_ELEMENT;
		if (state->element->type == PONG_TYPE_INVALID) {
			g_warning (_("Untyped Pong Element found"));
			pong_element_unref (state->element);
			state->element = NULL;
			break;
		}
		if ( ! state->element->conf_path) {
			g_warning (_("Unnamed Pong Element found"));
			pong_element_unref (state->element);
			state->element = NULL;
			break;
		}
		pong_file_add_element (state->gf, state->element);
		state->element = NULL;
		break;

	case INSIDE_BASECONFPATH:
		state->state = LOOKING_FOR_ELEMENT;
		g_free (state->gf->base_conf_path);
		state->gf->base_conf_path = g_strdup (state->buf->str);

		/* chop off trailing '/' */
		len = strlen (state->gf->base_conf_path);
		if (state->gf->base_conf_path[len-1] == '/')
			state->gf->base_conf_path[len-1] = '\0';

		/*FIXME: do further checking on path (gconf_valid_key??) */

		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_DIALOG_TITLE:
		state->state = LOOKING_FOR_ELEMENT;
		g_free (state->gf->dialog_title);
		state->gf->dialog_title = g_strdup (translate_string (state, state->buf->str));
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_AUTO_APPLY:
		state->state = LOOKING_FOR_ELEMENT;
		state->gf->auto_apply = pong_bool_from_string (state->buf->str);
		if ( ! state->gf->auto_apply) {
			g_warning (_("Non auto apply is NOT currently implemented"));
		}
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_REVERT:
		state->state = LOOKING_FOR_ELEMENT;
		state->gf->auto_apply = pong_bool_from_string (state->buf->str);
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_HELP_NAME:
		state->state = INSIDE_HELP;
		g_free (state->gf->help_name);
		state->gf->help_name = g_strdup (state->buf->str);
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_HELP_PATH:
		state->state = INSIDE_HELP;
		g_free (state->gf->help_path);
		state->gf->help_path = g_strdup (translate_string (state, state->buf->str));
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_HELP:
		state->state = LOOKING_FOR_ELEMENT;
		break;

	case INSIDE_LEVEL_NAME:
		state->state = INSIDE_LEVEL;
		g_free (state->level->name);
		state->level->name = g_strdup (state->buf->str);
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_LEVEL_BASECONFPATH:
		state->state = INSIDE_LEVEL;
		g_free (state->level->base_conf_path);
		state->level->base_conf_path = g_strdup (state->buf->str);
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_LEVEL:
		if (state->level->name && state->level->base_conf_path) {
			state->gf->levels = g_list_prepend (state->gf->levels, state->level);
		} else {
			g_warning (_("Fields missing from Level tag"));
			g_free (state->level->name);
			g_free (state->level->base_conf_path);
			g_free (state->level);
		}
		state->level = NULL;
		state->state = LOOKING_FOR_ELEMENT;
		break;

	case INSIDE_ELEMENT_TYPE:
		pong_element_set_type_string (state->element,
					      state->buf->str);

		/* stop reading characters */
		state->reading_characters = FALSE;

		state->state = INSIDE_ELEMENT;
		break;

	case INSIDE_ELEMENT_CONF_PATH:
		pong_element_set_conf_path (state->element,
					   state->buf->str);

		/* stop reading characters */
		state->reading_characters = FALSE;

		state->state = INSIDE_ELEMENT;
		break;

	case INSIDE_ELEMENT_WIDGET:
		pong_element_set_widget (state->element,
					state->buf->str);

		/* stop reading characters */
		state->reading_characters = FALSE;

		state->state = INSIDE_ELEMENT;
		break;

	case INSIDE_ELEMENT_DEFAULT:
		pong_element_set_def (state->element,
				      state->buf->str);

		/* stop reading characters */
		state->reading_characters = FALSE;

		state->state = INSIDE_ELEMENT;
		break;

	case INSIDE_ELEMENT_SPECIFIER:
		pong_element_set_specifier (state->element,
					   state->buf->str);

		/* stop reading characters */
		state->reading_characters = FALSE;

		state->state = INSIDE_ELEMENT;
		break;

	case INSIDE_SENSITIVITY:
		state->state = INSIDE_ELEMENT;
		pong_element_add_sensitivity (state->element, state->sens);
		state->sens = NULL;
		break;

	case INSIDE_SENSITIVITY_VALUE:
		pong_sensitivity_add_value_from_string (state->sens, 
							state->buf->str);
		state->state = INSIDE_SENSITIVITY;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_SENSITIVITY_CONNECTION:
		i = get_connection (state->buf->str);
		if (i == PONG_CONNECTION_INVALID) {
			g_warning (_("Unknown connection '%s' assuming OR"),
				  state->buf->str);
			i = PONG_CONNECTION_OR;
		}
		pong_sensitivity_set_connection (state->sens, i);
		state->state = INSIDE_SENSITIVITY;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_SENSITIVITY_COMPARISON:
		i = get_comparison (state->buf->str);
		if (i == PONG_COMP_INVALID) {
			g_warning (_("Unknown comparison '%s' assuming equals"),
				  state->buf->str);
			i = PONG_COMP_EQ;
		}
		pong_sensitivity_set_comparison (state->sens, i);
		state->state = INSIDE_SENSITIVITY;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_SENSITIVITY_SENSITIVE:
		pong_sensitivity_add_sensitive (state->sens, state->buf->str);
		state->state = INSIDE_SENSITIVITY;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_SENSITIVITY_INSENSITIVE:
		pong_sensitivity_add_insensitive (state->sens, state->buf->str);
		state->state = INSIDE_SENSITIVITY;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PANE:
		state->gf->panes =
			g_list_append (state->gf->panes, state->pane);
		state->pane = NULL;
		state->state = LOOKING_FOR_ELEMENT;
		break;

	case INSIDE_PANE_NAME:
		pong_pane_base_set_name (&state->pane->base, state->buf->str);
		state->state = INSIDE_PANE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PANE_LABEL:
		{
			const char *string = translate_string (state,
							       state->buf->str);
			pong_pane_base_set_label (&state->pane->base, string);

			state->state = INSIDE_PANE;
			/* stop reading characters */
			state->reading_characters = FALSE;
		}
		break;

	case INSIDE_PANE_LEVEL:
		pong_pane_base_add_level (&state->pane->base,
					  state->buf->str);
		state->state = INSIDE_PANE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GLADE:
		if (state->plug_widget->path == NULL ||
		    state->plug_widget->specifier == NULL) {
			g_warning (_("A glade widget needs to have 'File' and "
				     "'Widget' specified"));
		} else {
			if (state->group != NULL)
				pong_group_append_plug_widget (state->group, state->plug_widget);
			else
				pong_pane_append_plug_widget (state->pane, state->plug_widget);
		}
		pong_pane_base_unref (&state->plug_widget->base);
		state->plug_widget = NULL;
		state->state = state->state_before_plug;
		break;

	case INSIDE_GLADE_NAME:
		pong_pane_base_set_name (&state->plug_widget->base,
					 state->buf->str);
		state->state = INSIDE_GLADE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GLADE_LEVEL:
		pong_pane_base_add_level (&state->plug_widget->base,
					  state->buf->str);
		state->state = INSIDE_GLADE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GLADE_EXPAND:
		if (pong_bool_from_string (state->buf->str))
			pong_plug_widget_set_expand (state->plug_widget, TRUE);
		else
			pong_plug_widget_set_expand (state->plug_widget, FALSE);
		state->state = INSIDE_GLADE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GLADE_FILE:
		pong_plug_widget_set_path (state->plug_widget,
					   state->buf->str);
		state->state = INSIDE_GLADE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GLADE_WIDGET:
		pong_plug_widget_set_specifier (state->plug_widget,
						state->buf->str);
		state->state = INSIDE_GLADE;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_BONOBO:
		if (state->plug_widget->path == NULL) {
			g_warning (_("A bonobo widget needs to have 'Query' "
				     "specified"));
		} else {
			if (state->group != NULL)
				pong_group_append_plug_widget (state->group, state->plug_widget);
			else
				pong_pane_append_plug_widget (state->pane, state->plug_widget);
		}
		pong_pane_base_unref (&state->plug_widget->base);
		state->plug_widget = NULL;
		state->state = state->state_before_plug;
		break;

	case INSIDE_BONOBO_NAME:
		pong_pane_base_set_name (&state->plug_widget->base,
					state->buf->str);
		state->state = INSIDE_BONOBO;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_BONOBO_LEVEL:
		pong_pane_base_add_level (&state->plug_widget->base,
					  state->buf->str);
		state->state = INSIDE_BONOBO;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_BONOBO_EXPAND:
		if (pong_bool_from_string (state->buf->str))
			pong_plug_widget_set_expand (state->plug_widget, TRUE);
		else
			pong_plug_widget_set_expand (state->plug_widget, FALSE);
		state->state = INSIDE_BONOBO;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_BONOBO_QUERY:
		pong_plug_widget_set_path (state->plug_widget,
					   state->buf->str);
		state->state = INSIDE_BONOBO;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PLUG:
		if (state->plug_widget->plug_type == NULL ||
		    state->plug_widget->path == NULL) {
			g_warning (_("A plug widget needs to have 'Type' and "
				     "'Path' specified at least"));
		} else {
			if (state->group != NULL)
				pong_group_append_plug_widget (state->group, state->plug_widget);
			else
				pong_pane_append_plug_widget (state->pane, state->plug_widget);
		}
		pong_pane_base_unref (&state->plug_widget->base);
		state->plug_widget = NULL;
		state->state = state->state_before_plug;
		break;

	case INSIDE_PLUG_NAME:
		pong_pane_base_set_name (&state->plug_widget->base,
					state->buf->str);
		state->state = INSIDE_PLUG;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PLUG_LEVEL:
		pong_pane_base_add_level (&state->plug_widget->base,
					  state->buf->str);
		state->state = INSIDE_PLUG;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PLUG_EXPAND:
		if (pong_bool_from_string (state->buf->str))
			pong_plug_widget_set_expand (state->plug_widget, TRUE);
		else
			pong_plug_widget_set_expand (state->plug_widget, FALSE);
		state->state = INSIDE_PLUG;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PLUG_TYPE:
		pong_plug_widget_set_plug_type (state->plug_widget,
						state->buf->str);
		state->state = INSIDE_PLUG;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PLUG_PATH:
		pong_plug_widget_set_path (state->plug_widget,
					   state->buf->str);
		state->state = INSIDE_PLUG;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_PLUG_SPECIFIER:
		pong_plug_widget_set_specifier (state->plug_widget,
						state->buf->str);
		state->state = INSIDE_PLUG;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GROUP:
		pong_pane_append_group (state->pane, state->group);
		pong_pane_base_unref (&state->group->base);
		state->group = NULL;
		state->state = INSIDE_PANE;
		break;

	case INSIDE_GROUP_NAME:
		pong_pane_base_set_name (&state->group->base, state->buf->str);
		state->state = INSIDE_GROUP;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GROUP_LABEL:
		{
			const char *string = translate_string (state,
							       state->buf->str);
			pong_pane_base_set_label (&state->group->base, string);

			state->state = INSIDE_GROUP;
			/* stop reading characters */
			state->reading_characters = FALSE;
		}
		break;

	case INSIDE_GROUP_LEVEL:
		pong_pane_base_add_level (&state->group->base,
					  state->buf->str);
		state->state = INSIDE_GROUP;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_GROUP_EXPAND:
		if (pong_bool_from_string (state->buf->str))
			pong_group_set_expand (state->group, TRUE);
		else
			pong_group_set_expand (state->group, FALSE);
		state->state = INSIDE_GROUP;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_WIDGET:
		pong_group_append_widget (state->group, state->widget);
		pong_pane_base_unref (&state->widget->base);
		state->widget = NULL;
		state->state = INSIDE_GROUP;
		break;

	case INSIDE_WIDGET_NAME:
		pong_pane_base_set_name (&state->widget->base, state->buf->str);
		state->state = INSIDE_WIDGET;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_WIDGET_TOOLTIP:
		pong_widget_set_tooltip (state->widget, state->buf->str);
		state->state = INSIDE_WIDGET;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_WIDGET_LABEL:
		{
			const char *string = translate_string (state,
							       state->buf->str);
			pong_pane_base_set_label (&state->widget->base, string);

			state->state = INSIDE_WIDGET;
			/* stop reading characters */
			state->reading_characters = FALSE;
		}
		break;

	case INSIDE_WIDGET_LEVEL:
		pong_pane_base_add_level (&state->widget->base,
					  state->buf->str);
		state->state = INSIDE_WIDGET;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_WIDGET_EXPAND:
		if (pong_bool_from_string (state->buf->str))
			pong_widget_set_expand (state->widget, TRUE);
		else
			pong_widget_set_expand (state->widget, FALSE);
		state->state = INSIDE_WIDGET;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_WIDGET_TYPE:
		if (state->bogus_types)
			/* bogus types means we aren't trying to
			   parse types, useful for some parsing, where
			   we don't care, about types */
			pong_widget_set_widget_type (state->widget,
						     PONG_INVALIDWIDGET);
		else
			pong_widget_set_widget_type_string (state->widget,
							    state->buf->str);
		state->state = INSIDE_WIDGET;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_ARGUMENT:
		if ( ! state->arg_name ||
		     ! state->arg_value) {
			g_warning (_("Argument requires both name and value"));
			g_free (state->arg_name);
			g_free (state->arg_value);
			state->arg_name = NULL;
			state->arg_value = NULL;
			state->state = INSIDE_WIDGET;
			break;
		}

		if (state->arg_translatable) {
			const char *string = translate_string (state,
							       state->arg_value);
			pong_widget_add_argument (state->widget,
						  state->arg_name,
						  string,
						  state->arg_translatable);
		} else {
			pong_widget_add_argument (state->widget,
						 state->arg_name,
						 state->arg_value,
						 state->arg_translatable);
		}
		g_free (state->arg_name);
		g_free (state->arg_value);
		state->arg_name = NULL;
		state->arg_value = NULL;
		state->state = INSIDE_WIDGET;
		break;

	case INSIDE_ARGUMENT_NAME:
		g_free (state->arg_name);
		state->arg_name = g_strdup (state->buf->str);
		state->state = INSIDE_ARGUMENT;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_ARGUMENT_VALUE:
		g_free (state->arg_value);
		state->arg_value = g_strdup (state->buf->str);
		state->state = INSIDE_ARGUMENT;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_ARGUMENT_TRANSLATABLE:
		if (pong_bool_from_string (state->buf->str))
			state->arg_translatable = TRUE;
		else
			state->arg_translatable = FALSE;
		state->state = INSIDE_ARGUMENT;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_OPTION:
		if ( ! state->opt_value) {
			g_warning (_("Option requires a value"));
			g_free (state->opt_label);
			state->opt_label = NULL;
			state->opt_value = NULL;
			state->state = INSIDE_WIDGET;
			break;
		}

		if (state->opt_label != NULL) {
			const char *string = translate_string (state,
							       state->opt_label);
			pong_widget_add_option (state->widget,
						string,
						state->opt_value);
		} else {
			pong_widget_add_option (state->widget, NULL,
					       state->opt_value);
		}
		state->state = INSIDE_WIDGET;
		break;

	case INSIDE_OPTION_LABEL:
		g_free (state->opt_label);
		state->opt_label = g_strdup (state->buf->str);
		state->state = INSIDE_OPTION;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	case INSIDE_OPTION_VALUE:
		g_free (state->opt_value);
		state->opt_value = g_strdup (state->buf->str);
		state->state = INSIDE_OPTION;
		/* stop reading characters */
		state->reading_characters = FALSE;
		break;

	default:
		g_warning ("UHH? (state %d)", state->state);
		break;
	}
}


/* stolen from the SAX interface tutorial */
static int
myXmlSAXParseFile (xmlSAXHandlerPtr sax,
		   gpointer user_data,
		   const char *filename)
{
	int ret = 0;                                        
	xmlParserCtxtPtr ctxt;                              

	ctxt = xmlCreateFileParserCtxt (filename);    
	if (ctxt == NULL) return -1;                 
	ctxt->sax = sax;                                    
	ctxt->userData = user_data;    

	xmlParseDocument (ctxt);      

	if (ctxt->wellFormed)        
		ret = 0;                                   
	else                                               
		ret = -1;          
	if (sax != NULL)                                   
		ctxt->sax = NULL;
	xmlFreeParserCtxt (ctxt);     

	return ret;
}

static int
myXmlSAXParseMemory (xmlSAXHandlerPtr sax,
		     gpointer user_data,
		     const char *buffer,
		     int size)
{
	int ret = 0;                                        
	xmlParserCtxtPtr ctxt;                              

	/* XXX: casting to (char *) should be safe */
	ctxt = xmlCreateMemoryParserCtxt ((char *)buffer, size);    
	if (ctxt == NULL) return -1;                 
	ctxt->sax = sax;                                    
	ctxt->userData = user_data;    

	xmlParseDocument (ctxt);      

	if (ctxt->wellFormed)        
		ret = 0;                                   
	else                                               
		ret = -1;          
	if (sax != NULL)                                   
		ctxt->sax = NULL;
	xmlFreeParserCtxt (ctxt);     

	return ret;
}

static void
pong_file_add_element (PongFile *gf, PongElement *ge)
{
	PongElement *old;
	old = g_hash_table_lookup (gf->element_hash,
				   ge->conf_path);
	if (old != NULL) {
		g_hash_table_remove (gf->element_hash,
				     ge->conf_path);
		pong_element_unref (old);
	}
	g_hash_table_insert (gf->element_hash,
			     ge->conf_path, ge);
}

static void
unref_elements (gpointer key, gpointer data, gpointer user_data)
{
	PongElement *ge = data;
	pong_element_unref (ge);
}

static void
pong_file_destroy (PongFile *gf)
{
	GList *li;
	g_hash_table_foreach (gf->element_hash,
			      unref_elements, NULL);
	g_hash_table_destroy (gf->element_hash);
	gf->element_hash = NULL;

	g_free (gf->file_name);
	gf->file_name = NULL;

	g_free (gf->base_conf_path);
	gf->base_conf_path = NULL;

	g_free (gf->dialog_title);
	gf->dialog_title = NULL;

	g_free (gf->help_name);
	gf->help_name = NULL;
	g_free (gf->help_path);
	gf->help_path = NULL;

	g_list_foreach (gf->panes, (GFunc)pong_pane_base_unref, NULL);
	gf->panes = NULL;

	for (li = gf->levels; li != NULL; li = li->next) {
		PongLevel *level = li->data;
		li->data = NULL;
		g_free (level->name);
		g_free (level->base_conf_path);
		g_free (level);
	}
	g_list_free (gf->levels);
	gf->levels = NULL;

	g_free (gf);
}

static void
clear_state (State *state)
{
	if (state->gf != NULL)
		pong_file_unref (state->gf);
	state->gf = NULL;

	if (state->sens != NULL)
		pong_sensitivity_destroy (state->sens);
	state->sens = NULL;

	if (state->element != NULL)
		pong_element_unref (state->element);
	state->element = NULL;

	if (state->pane != NULL)
		pong_pane_base_unref (&state->pane->base);
	state->pane = NULL;

	if (state->group != NULL)
		pong_pane_base_unref (&state->group->base);
	state->group = NULL;

	if (state->widget != NULL)
		pong_pane_base_unref (&state->widget->base);
	state->widget = NULL;

	if (state->plug_widget != NULL)
		pong_pane_base_unref (&state->plug_widget->base);
	state->plug_widget = NULL;

	g_free (state->gettext_domain);
	state->gettext_domain = NULL;

	pong_translation_unref (state->trans);
	state->trans = NULL;

	g_free (state->arg_name);
	state->arg_name = NULL;
	g_free (state->arg_value);
	state->arg_value = NULL;

	g_free (state->opt_label);
	state->opt_label = NULL;
	g_free (state->opt_value);
	state->opt_value = NULL;

	if (state->level != NULL)
		pong_level_destroy (state->level);
	state->level = NULL;
}

PongFile *
pong_file_load_xml (const char *file,
		    gboolean no_i18n,
		    gboolean use_gettext,
		    const char *gettext_domain,
		    gboolean bogus_types)
{
	State state = {0};
	PongFile *gf;
	char *found_file;

	g_return_val_if_fail (file != NULL, NULL);

	if ( ! parser_inited)
		parser_init ();

	gf = g_hash_table_lookup (file_hash, file);
	if (gf != NULL) {
		return pong_file_ref (gf);
	}

	found_file = pong_find_file (file, pong_directories);

	if (found_file == NULL) {
		g_warning (_("Cannot find file '%s'"), file);
		return NULL;
	}

	gf = g_hash_table_lookup (file_hash, found_file);
	if (gf != NULL) {
		g_free (found_file);
		return pong_file_ref (gf);
	}

	if (no_i18n)
		state.use_gettext = FALSE;
	else
		state.use_gettext = use_gettext;
	state.gettext_domain = g_strdup (gettext_domain);
	state.bogus_types = bogus_types;

	if ( ! use_gettext &&
	     ! no_i18n) {
		state.trans = pong_translation_make (found_file);
	}

	/* Note, if no_i18n is TRUE, then use_gettext will be FALSE
	 * and trans will be NULL, thus getting no translations */

	state.gf = g_new0 (PongFile, 1);
	state.gf->base_conf_path = g_strdup ("/");
	state.gf->cached = FALSE;
	state.gf->element_hash = g_hash_table_new (g_str_hash, g_str_equal);
	state.gf->auto_apply = TRUE;
	state.gf->revert = TRUE;

	/* take over owenership of found_file */
	state.gf->file_name = found_file;

	if (myXmlSAXParseFile (&handler, &state, found_file) < 0) {
		pong_file_destroy (state.gf);
		gf = NULL;
	} else {
		g_hash_table_insert (file_hash, g_strdup (found_file), state.gf);
		/* if cached is on, the object will not get destroyed on the
		 * last unref, but it will stick around in the hash */
		state.gf->cached = TRUE;
		/* ref for the user */
		gf = pong_file_ref (state.gf);
		state.gf = NULL;
	}

	clear_state (&state);

	return gf;
}

PongFile *
pong_file_load_xml_from_memory (const char *buffer,
				int size,
				gboolean use_gettext,
				const char *gettext_domain,
				PongTranslation *trans,
				gboolean bogus_types)
{
	State state = {0};

	g_return_val_if_fail (buffer != NULL, NULL);
	g_return_val_if_fail (size > 0, NULL);

	if ( ! parser_inited)
		parser_init ();

	state.use_gettext = use_gettext;
	state.bogus_types = bogus_types;
	state.gettext_domain = g_strdup (gettext_domain);

	if ( ! use_gettext)
		state.trans = pong_translation_ref (trans);

	state.gf = g_new0 (PongFile, 1);
	state.gf->cached = FALSE;
	state.gf->element_hash = g_hash_table_new (g_str_hash, g_str_equal);

	if (myXmlSAXParseMemory (&handler, &state, buffer, size) < 0) {
		clear_state (&state);
		return NULL;
	} else {
		PongFile *gf = state.gf;
		state.gf = NULL;
		clear_state (&state);
		return gf;
	}
}

PongFile *
pong_file_ref (PongFile *gf)
{
	g_return_val_if_fail (gf != NULL, NULL);

	gf->refcount ++;

	return gf;
}

void
pong_file_unref (PongFile *gf)
{
	g_return_if_fail (gf != NULL);

	gf->refcount --;

	if (gf->refcount == 0 && !gf->cached)
		pong_file_destroy (gf);
}

static gboolean
kill_unused_pongfiles (gpointer key, gpointer data, gpointer user_data)
{
	PongFile *gf = data;
	if (gf->cached && gf->refcount <= 0) {
		gf->cached = FALSE;
		pong_file_destroy (gf);

		/* free the filename */
		g_free (data);
		return TRUE;
	}
	return FALSE;
}

void
pong_parser_free_unused_data (void)
{
	if (parser_inited) {
		g_hash_table_foreach_remove (file_hash, kill_unused_pongfiles, NULL);

		if (g_hash_table_size (file_hash) == 0) {
			g_hash_table_destroy (file_hash);
			file_hash = NULL;
			parser_inited = FALSE;
		}
	}

	if (comparison_hash != NULL) {
		g_hash_table_destroy (comparison_hash);
		comparison_hash = NULL;
	}
}

static void
add_elements_to_list (gpointer key, gpointer data, gpointer user_data)
{
	GList **list = user_data;
	*list = g_list_prepend (*list, key);
}

/* list of internal strings, DON'T FREE THEM, just free the list */
GList *
pong_file_get_conf_paths (const PongFile *gf)
{
	GList *list = NULL;
	g_return_val_if_fail (gf != NULL, NULL);

	g_hash_table_foreach (gf->element_hash, add_elements_to_list, &list);

	return list;
}

PongElement *
pong_file_lookup_element (const PongFile *gf, const char *conf_path)
{
	g_return_val_if_fail (gf != NULL, NULL);
	g_return_val_if_fail (conf_path != NULL, NULL);

	return g_hash_table_lookup (gf->element_hash, conf_path);
}

static void
foreach_elements (gpointer key, gpointer data, gpointer user_data)
{
	gpointer *fedata = user_data;
	PongElementFunc func = fedata[0];

	func ((PongElement *)data, fedata[1]);
}

void
pong_file_foreach_element (PongFile *gf, PongElementFunc func, gpointer data)
{
	gpointer fedata[2];

	g_return_if_fail (gf != NULL);
	g_return_if_fail (func != NULL);

	fedata[0] = func;
	fedata[1] = data;

	g_hash_table_foreach (gf->element_hash, foreach_elements, fedata);
}

PongLevel *
pong_level_new (void)
{
	PongLevel *level = g_new0 (PongLevel, 1);
	return level;
}

void
pong_level_destroy (PongLevel *level)
{
	g_return_if_fail (level != NULL);

	g_free (level->name);
	level->name = NULL;
	g_free (level->base_conf_path);
	level->base_conf_path = NULL;
	g_free (level);
}


GList *
pong_file_get_levels (PongFile *gf)
{
	GList *li, *list = NULL;

	g_return_val_if_fail (gf != NULL, NULL);

	for (li = gf->levels; li != NULL; li = li->next) {
		PongLevel *level = li->data;
		list = g_list_prepend (list, level->name);
	}

	return list;
}

const char *
pong_file_get_level_base_conf_path (PongFile *gf, const char *level_name)
{
	GList *li;

	g_return_val_if_fail (gf != NULL, NULL);
	g_return_val_if_fail (level_name != NULL, NULL);

	for (li = gf->levels; li != NULL; li = li->next) {
		PongLevel *level = li->data;
		if (strcmp (level_name, level->name) == 0)
			return level->base_conf_path;
	}
	return NULL;
}
