/* PonG: Translation routines
 * Author: George Lebl
 * (c) 2000 Eazel, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include <gnome.h>
#include <stdlib.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>
#include <gnome-xml/xmlmemory.h>

#include "pong-strings.h"
#include "pongutil.h"

#include "pongtrans.h"

/* A private opaque structure */
struct _PongTranslation {
	int ref_count;
	GHashTable *hash_table;
};


/**
 * pong_make_translation_file:
 * @pong_filename: The .pong filename
 * @locale: The locale
 *
 * Description:  Makes a translation filename from a pong filename.
 * This is done by removing the ".pong" ending from the filename and
 * appending .<locale>.pong_po
 **/
char *
pong_make_translation_file (const char *pong_filename,
			    const char *locale)
{
	char *ns, *p;

	g_return_val_if_fail (pong_filename != NULL, NULL);
	g_return_val_if_fail (locale != NULL, NULL);

	ns = g_new (char, strlen (pong_filename) +
		    strlen (locale) + strlen ("..pong_po") + 1);

	strcpy (ns, pong_filename);

	p = strrchr (ns, '.');
	if (p != NULL &&
	    strcmp (p, ".pong") == 0)
		*p = '\0';

	strcat (ns, ".");
	strcat (ns, locale);
	strcat (ns, ".pong_po");

	return ns;
}


PongTranslation *
pong_translation_make (const char *pong_filename)
{
	const GList *locales;
	PongTranslation *trans = NULL;

	g_return_val_if_fail (pong_filename != NULL, NULL);

	locales = gnome_i18n_get_language_list ("LC_MESSAGES");

	while (locales != NULL &&
	       trans == NULL) {
		const char *locale;
		char *file;

		/* set locale */
		locale = locales->data;

		/* iterate to next */
		locales = locales->next;

		if (strcmp (locale, "C") == 0)
			continue;

		file = pong_make_translation_file (pong_filename, locale);

		trans = pong_translation_from_file (file);

		g_free (file);
	}

	return trans;
}

static void
read_translation (PongTranslation *trans, xmlNodePtr parent)
{
	xmlNodePtr node;

	char *string = NULL;
	char *translation = NULL;

	for (node = parent->childs; node != NULL; node = node->next) {
		if (pong_strcasecmp_no_locale (node->name, PONG_S_String) == 0) {
			if (string != NULL)
				xmlFree (string);
			string = xmlNodeGetContent (node);
		} else if (pong_strcasecmp_no_locale (node->name, PONG_S_Translated) == 0) {
			if (translation != NULL)
				xmlFree (translation);
			translation = xmlNodeGetContent (node);
		}
	}

	if (string != NULL &&
	    string[0] != '\0' &&
	    translation != NULL &&
	    translation[0] != '\0') {
		gpointer old_key, old_val;

		if (g_hash_table_lookup_extended (trans->hash_table, string,
						  &old_key, &old_val)) {
			g_free (old_val);
			g_hash_table_insert (trans->hash_table, old_key,
					     g_strdup (translation));
		} else {
			g_hash_table_insert (trans->hash_table,
					     g_strdup (string),
					     g_strdup (translation));
		}
	}

	if (string != NULL)
		xmlFree (string);

	if (translation != NULL)
		xmlFree (translation);
}

static void
read_translations (PongTranslation *trans, xmlDocPtr doc)
{
	xmlNodePtr node;

	for (node = doc->root->childs; node != NULL; node = node->next) {
		if (pong_strcasecmp_no_locale (node->name, PONG_S_Translation) == 0) {
			read_translation (trans, node);
		}
	}
}

static PongTranslation *
pong_translation_new (void)
{
	PongTranslation *trans;

	trans = g_new0 (PongTranslation, 1);
	trans->ref_count = 1;
	trans->hash_table = g_hash_table_new (g_str_hash, g_str_equal);

	return trans;
}

PongTranslation *
pong_translation_from_file (const char *file)
{
	PongTranslation *trans = NULL;

	if (file != NULL &&
	    g_file_exists (file)) {
		xmlDocPtr doc;

		doc = xmlParseFile (file);

		if (doc != NULL &&
		    doc->root != NULL &&
		    doc->root->name != NULL &&
		    pong_strcasecmp_no_locale(doc->root->name, PONG_S_PongTranslations) == 0) {
			trans = pong_translation_new ();
			read_translations (trans, doc);
		}

		xmlFreeDoc(doc);
	}

	return trans;
}

PongTranslation *
pong_translation_from_memory (const char *buffer, int size)
{
	xmlDocPtr doc;
	PongTranslation *trans = NULL;

	g_return_val_if_fail (buffer != NULL, NULL);
	g_return_val_if_fail (size > 0, NULL);

	/* Casting is OK here */
	doc = xmlParseMemory ((char *)buffer,
			      size);

	if (doc != NULL &&
	    doc->root != NULL &&
	    doc->root->name != NULL &&
	    pong_strcasecmp_no_locale(doc->root->name, PONG_S_PongTranslations) == 0) {
		trans = pong_translation_new ();
		read_translations (trans, doc);
	}

	xmlFreeDoc(doc);

	return trans;
}

const char *
pong_translation_translate (PongTranslation *trans, const char *string)
{
	if (trans != NULL) {
		const char *ret;

		ret = g_hash_table_lookup (trans->hash_table, string);

		if (ret != NULL)
			string = ret;
	}
	return string;
}

PongTranslation *
pong_translation_ref (PongTranslation *trans)
{
	if (trans != NULL) {
		trans->ref_count ++;
	}
	return trans;
}

static void
free_table (gpointer key, gpointer data, gpointer user_data)
{
	g_free (key);
	g_free (data);
}

void
pong_translation_unref (PongTranslation *trans)
{
	if (trans != NULL) {
		trans->ref_count --;

		if (trans->ref_count == 0) {
			g_hash_table_foreach (trans->hash_table, free_table, NULL);
			g_hash_table_destroy (trans->hash_table);
			trans->hash_table = NULL;

			g_free (trans);
		}
	}
}

